#include <osfcn.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <errno.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <time.h>
#include "adosfilsys.h"


#define max(a, b) ((a)>(b)?(a):(b))
#define min(a, b) ((a)<(b)?(a):(b))


void panic(char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);
    (void)vfprintf(stderr, fmt, ap);
    va_end(ap);
    putc('\n', stderr);
    exit(-1);
}


void *Malloc(unsigned nbytes)
{
    void *p = malloc(nbytes);
    if (!p)
	panic("malloc failure: %s", strerror(errno));
    return p;
}


void *Free(void *p)
{
    free(p);
}


int fd;
int mode = O_RDONLY;


struct buf
{
    int flags;
    int ref;
    struct buf *next;
    int blkno;
    unsigned char *addr;
};

#define B_VALID	1
#define B_MAP	2

struct buf *cache;

#define BlocksPerPage 8			/* Map this many blocks at a time */
#define PAGESIZE (BLOCKSIZE*BlocksPerPage)

struct buf *bread(int blkno)
{
    struct buf *b;
    unsigned char *addr;
    int i, baseblock;
    unsigned long base;

    for ( b=cache ; b ; b=b->next )
	if (b->blkno == blkno)
	    return b;

    baseblock = blkno - (blkno%BlocksPerPage);
    base = (baseblock)*BLOCKSIZE;
    addr = (unsigned char *)mmap((caddr_t)0, PAGESIZE,
				 ((mode==O_RDWR) ?
				  PROT_READ|PROT_WRITE : PROT_READ),
				 MAP_SHARED, fd, base);
    if ((addr == (unsigned char *)-1))
    {
	addr = Malloc(BLOCKSIZE);
	b = Malloc(sizeof *b);
	lseek(fd, blkno*BLOCKSIZE, 0);
	if (read(fd, addr, BLOCKSIZE) != BLOCKSIZE)
	    panic("read of block %d failed: %s", blkno, strerror(errno));
	b->flags = B_VALID;
	b->ref = 1;
	b->blkno = blkno;
	b->addr = addr;
	b->next = cache;
	cache = b;
	return b;
    }

    b = Malloc(BlocksPerPage * sizeof *b);
    for ( i=0 ; i<BlocksPerPage ; ++ i )
    {
	b[i].flags = B_MAP|B_VALID;
	b[i].ref = 0;
	b[i].blkno = baseblock + i;
	b[i].addr = addr + BLOCKSIZE * i;
	b[i].next = b+i+1;
    }
    b[BlocksPerPage-1].next = cache;
    cache = b;

    b += (blkno - baseblock);
    ++b->ref;
    return b;
}

void brelse(struct buf *b)
{
    --b->ref;
}


clock_t cvtadostime(struct adostime *at, clock_t tzoff)
{
    tzoff += TIME_OFFSET;
    tzoff += at->days * 86400;
    tzoff += at->mins * 60;
    tzoff += at->ticks / TICKS_PER_SECOND;

    return tzoff;
}


static int ToUpper(unsigned char c)
{
    return (c>='a' && c<='z') ? c+'A'-'a' : c;
}

int adoshash(unsigned char *name, int len)
{
    int i, val=len;
    for ( i=0 ; i<len ; ++i )
	val = ((val*13) + ToUpper(*name++)) & 0x7ff;
    return val % 72;
}


int FFS;
int totalsize = 1760;
int beginreserved = 2;
int endreserved = 0;
int rootblkno;


struct filsys *filsys;


void dumpdir(struct dinode *dir, int lev)
{
    int i, bn;
    struct buf *b;
    struct dinode *dinode;
    clock_t timestamp;

    for ( i=0 ; i<NADDR ; ++i )
	for ( bn=dir->addr[i] ; bn ; bn=dinode->hashchain )
	{
	    b = bread(bn);
	    dinode = (struct dinode *)b->addr;
	    timestamp = cvtadostime(&dinode->mtime, 0L);
	    printf("%*s%2d/%-4d  %.24s %8ld `%.*s'\n",
		   lev, "",
		   dinode->type, dinode->stype,
		   ctime(&timestamp), dinode->size,
		   dinode->namelen, dinode->name);
	    if (dinode->type == T_SHORT &&
		dinode->stype == ST_USERDIR)
		dumpdir(dinode, lev+2);
	}
}


int main(int argc, char *argv[])
{
    struct buf *b;
    clock_t timestamp;

    if (argc < 2)
	panic("Usage: %s device [rootblock]", argv[0]);

    fd = open(argv[1], mode);
    if (fd < 0)
	panic("can't open %s: %s", argv[1], strerror(errno));

    if (argc > 2)
	rootblkno = strtol(argv[2], (char **)0, 0);
    else
	rootblkno = (totalsize-1 + beginreserved) / 2;

    b = bread(0);
    if (!memcmp(b->addr, "DOS\1", 4))
	FFS=1;
    else
	if (memcmp(b->addr, "DOS\0", 4))
	    fprintf(stderr, "Warning: assuming filesystem is DOS\\0\n");
    brelse(b);

    b = bread(rootblkno);
    filsys = (struct filsys *)b->addr;

    if (filsys->type != T_SHORT ||
	filsys->stype != ST_ROOT)
	panic("Invalid type/stype: %d/%d\n", filsys->type, filsys->stype);

    timestamp = cvtadostime(&filsys->mtime, 0L);
    printf("date: %.24s, fsname: `%.*s'\n",
	   ctime(&timestamp),
	   filsys->namelen, filsys->name);

    printf("Root directory contains:\n");
    dumpdir((struct dinode *)filsys, 2);

    return 0;
}
