#include <osfcn.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <errno.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <time.h>
#include "adosfilsys.h"


#define max(a, b) ((a)>(b)?(a):(b))
#define min(a, b) ((a)<(b)?(a):(b))


void panic(char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);
    (void)vfprintf(stderr, fmt, ap);
    va_end(ap);
    putc('\n', stderr);
    exit(-1);
}


void *Malloc(unsigned nbytes)
{
    void *p = malloc(nbytes);
    if (!p)
	panic("malloc failure: %s", strerror(errno));
    return p;
}


void *Free(void *p)
{
    free(p);
}


int fd;
int mode = O_RDONLY;


struct buf
{
    int flags;
    int ref;
    struct buf *next;
    int blkno;
    unsigned char *addr;
};

#define B_VALID	1
#define B_MAP	2

struct buf *cache;

#define BlocksPerPage 8			/* Map this many blocks at a time */
#define PAGESIZE (BLOCKSIZE*BlocksPerPage)

struct buf *bread(int blkno)
{
    struct buf *b;
    unsigned char *addr;
    int i, baseblock;
    unsigned long base;

    for ( b=cache ; b ; b=b->next )
	if (b->blkno == blkno)
	    return b;

    baseblock = blkno - (blkno%BlocksPerPage);
    base = (baseblock)*BLOCKSIZE;
    addr = (unsigned char *)mmap((caddr_t)0, PAGESIZE,
				 ((mode==O_RDWR) ?
				  PROT_READ|PROT_WRITE : PROT_READ),
				 MAP_SHARED, fd, base);
    if ((addr == (unsigned char *)-1))
    {
	addr = Malloc(BLOCKSIZE);
	b = Malloc(sizeof *b);
	lseek(fd, blkno*BLOCKSIZE, 0);
	if (read(fd, addr, BLOCKSIZE) != BLOCKSIZE)
	    panic("read of block %d failed: %s", blkno, strerror(errno));
	b->flags = B_VALID;
	b->ref = 1;
	b->blkno = blkno;
	b->addr = addr;
	b->next = cache;
	cache = b;
	return b;
    }

    b = Malloc(BlocksPerPage * sizeof *b);
    for ( i=0 ; i<BlocksPerPage ; ++ i )
    {
	b[i].flags = B_MAP|B_VALID;
	b[i].ref = 0;
	b[i].blkno = baseblock + i;
	b[i].addr = addr + BLOCKSIZE * i;
	b[i].next = b+i+1;
    }
    b[BlocksPerPage-1].next = cache;
    cache = b;

    b += (blkno - baseblock);
    ++b->ref;
    return b;
}

void brelse(struct buf *b)
{
    --b->ref;
}


clock_t cvtadostime(struct adostime *at, clock_t tzoff)
{
    tzoff += TIME_OFFSET;
    tzoff += at->days * 86400;
    tzoff += at->mins * 60;
    tzoff += at->ticks / TICKS_PER_SECOND;

    return tzoff;
}


static int ToUpper(unsigned char c)
{
    return (c>='a' && c<='z') ? c+'A'-'a' : c;
}

int adoshash(unsigned char *name, int len)
{
    int i, val=len;
    for ( i=0 ; i<len ; ++i )
	val = ((val*13) + ToUpper(*name++)) & 0x7ff;
    return val % 72;
}


int FFS;
int totalsize = 1760;
int beginreserved = 2;
int endreserved = 0;
int rootblkno;


struct filsys *filsys;


void dumpdir(struct dinode *dir)
{
    int i, bn;
    struct buf *b;
    struct dinode *dinode;
    clock_t timestamp;

    for ( i=0 ; i<NADDR ; ++i )
	for ( bn=dir->addr[i] ; bn ; bn=dinode->hashchain )
	{
	    b = bread(bn);
	    dinode = (struct dinode *)b->addr;
	    printf("%.*s\n", dinode->namelen, dinode->name);
	}
}


void dumpfile(struct dinode *dinode)
{
    int bn, blkno, firstindir=0;
    struct buf *b=0, *ib=0;
    struct datablock *db;
    struct indirblock *indir=(struct indirblock *)dinode;
    unsigned long size, n;

    bn=0;
    for ( size=dinode->size ; size>0 ; size-=n )
    {
	while (bn >= firstindir+NADDR)
	{
	    int nextindir = indir->next;
	    if (!nextindir)
		panic("block %d doesn't locate block %d", indir->key, bn);
	    if (ib)
		brelse(ib);
	    ib = bread(nextindir);
	    indir = (struct indirblock *)ib->addr;
	    firstindir += NADDR;
	}
	blkno = indir->addr[NADDR-1-(bn-firstindir)];

	if (!blkno)
	    panic("block %d of file has zero block", bn);
	if (b)
	    brelse(b);
	b = bread(blkno);
	if (FFS)
	{
	    n = min(BLOCKSIZE, size);
	    write(1, b->addr, n);
	}
	else
	{
	    db = (struct datablock *)b->addr;
	    n = min(sizeof db->data, size);
	    write(1, db->data, n);
	}
	++bn;
    }
}


struct dinode *nami(struct dinode *cd, unsigned char *path)
{
    struct buf *b;
    unsigned char *rest;
    int i, bn;
    struct dinode *dinode;
    int len;

    if (!*path)
	return cd;

    if (*path == '/')
    {
	if (!cd->parent)
	    return 0;
	b = bread(cd->parent);
	cd = nami((struct dinode *)b->addr, path+1);
	return cd;
    }

    for ( rest=path ; *rest ; ++rest )
	if (*rest == '/')
	{
	    *rest++ = 0;
	    break;
	}

    len = strlen(path);
    i = adoshash(path, len);
    for ( bn=cd->addr[i] ; bn ; bn=dinode->hashchain )
    {
	b = bread(bn);
	dinode = (struct dinode *)b->addr;
	if (dinode->namelen == len &&
	    !memcmp(dinode->name, path, len))
	    return nami(dinode, rest);
    }

    return 0;
}


int main(int argc, char *argv[])
{
    struct buf *b;
    struct dinode *dinode;
    clock_t timestamp;

    if (argc < 3)
	panic("Usage: %s device pathname [rootblock]", argv[0]);

    fd = open(argv[1], mode);
    if (fd < 0)
	panic("can't open %s: %s", argv[1], strerror(errno));

    if (argc > 3)
	rootblkno = strtol(argv[3], (char **)0, 0);
    else
	rootblkno = (totalsize-1 + beginreserved) / 2;

    b = bread(0);
    if (!memcmp(b->addr, "DOS\1", 4))
	FFS=1;
    else
	if (memcmp(b->addr, "DOS\0", 4))
	    fprintf(stderr, "Warning: assuming filesystem is DOS\\0\n");
    brelse(b);

    b = bread(rootblkno);
    filsys = (struct filsys *)b->addr;

    if (filsys->type != T_SHORT ||
	filsys->stype != ST_ROOT)
	panic("Invalid type/stype: %d/%d\n", filsys->type, filsys->stype);

    dinode = nami((struct dinode *)filsys, (unsigned char *)argv[2]);

    if (!dinode)
	panic("`%s' not found", argv[2]);

    timestamp = cvtadostime(&dinode->mtime, 0L);
    fprintf(stderr, "%2d/%-4d  %.24s  %8ld `%.*s'\n",
	    dinode->type, dinode->stype,
	    ctime(&timestamp), dinode->size,
	    dinode->namelen, dinode->name);

    if (dinode->type == T_SHORT &&
	(dinode->stype == ST_ROOT ||
	 dinode->stype == ST_USERDIR))
	dumpdir(dinode);
    else if (dinode->type == T_SHORT &&
	     dinode->stype == ST_FILE)
	dumpfile(dinode);
    else
	panic("Unknown type/stype: %d/%d\n", dinode->type, dinode->stype);

    return 0;
}
